/*
 * Copyright © 2004-2020 L2J Server
 * 
 * This file is part of L2J Server.
 * 
 * L2J Server is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * L2J Server is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package com.l2jserver.gameserver.data.xml.impl;

import java.util.HashMap;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;

import com.l2jserver.gameserver.model.base.ClassId;
import com.l2jserver.gameserver.model.base.ClassInfo;
import com.l2jserver.gameserver.util.IXmlReader;

/**
 * Loads the the list of classes and it's info.
 * @author Zoey76
 */
public final class ClassListData implements IXmlReader {
	
	private static final Logger LOG = LoggerFactory.getLogger(ClassListData.class);
	
	private final Map<ClassId, ClassInfo> _classData = new HashMap<>();
	
	protected ClassListData() {
		load();
	}
	
	@Override
	public void load() {
		_classData.clear();
		parseDatapackFile("data/stats/chars/classList.xml");
		LOG.info("Loaded {} class data.", _classData.size());
	}
	
	@Override
	public void parseDocument(Document doc) {
		for (Node n = doc.getFirstChild(); n != null; n = n.getNextSibling()) {
			if ("list".equals(n.getNodeName())) {
				for (Node d = n.getFirstChild(); d != null; d = d.getNextSibling()) {
					final NamedNodeMap attrs = d.getAttributes();
					if ("class".equals(d.getNodeName())) {
						Node attr = attrs.getNamedItem("classId");
						final ClassId classId = ClassId.getClassId(parseInteger(attr));
						attr = attrs.getNamedItem("name");
						final String className = attr.getNodeValue();
						attr = attrs.getNamedItem("parentClassId");
						final ClassId parentClassId = (attr != null) ? ClassId.getClassId(parseInteger(attr)) : null;
						_classData.put(classId, new ClassInfo(classId, className, parentClassId));
					}
				}
			}
		}
	}
	
	/**
	 * Gets the class list.
	 * @return the complete class list
	 */
	public Map<ClassId, ClassInfo> getClassList() {
		return _classData;
	}
	
	/**
	 * Gets the class info.
	 * @param classId the class ID
	 * @return the class info related to the given {@code classId}
	 */
	public ClassInfo getClass(ClassId classId) {
		return _classData.get(classId);
	}
	
	/**
	 * Gets the class info.
	 * @param classId the class Id as integer
	 * @return the class info related to the given {@code classId}
	 */
	public ClassInfo getClass(int classId) {
		final ClassId id = ClassId.getClassId(classId);
		return (id != null) ? _classData.get(id) : null;
	}
	
	public static ClassListData getInstance() {
		return SingletonHolder.INSTANCE;
	}
	
	private static class SingletonHolder {
		protected static final ClassListData INSTANCE = new ClassListData();
	}
}